#pragma once

#include "OptimusMini.h"

// **************** Features to be enabled or disabled in compile time ****************

// Enables "block" parameter to make caller thread wait until command execution completed.
// Device communication is queued in separate thread, but sometime we need to wait for
// real command execution (to get device ID or wait until displays are turned of)
#define FEATURE_BLOCKING

// Enables device "ping" every second. If the device receive nothing in about 12 seconds, 
// it turns displays off, which is useful after reset or if software terminated abnormally. 
#define FEATURE_PING

// If for some reasons no command was send for more then 8 seconds, device might switch off.
// Switching it on automaticaly in that case.
#define FEATURE_AUTO_ON

// Enables ChangeLayout method for fast layout change. If we want to change all three images, we want it fast. 
// Thus we first upload all three images to device, and only after that executing commands to show them. 
// It does not influence total time required, but pictures will be changed not one by one, but altogether.
#define FEATURE_LAYOUT

// Enables RepaintPictures method. Useful to restore images after device is unplugged and plugged back.
#define FEATURE_REPAINT

// EventHandler can be registered for to customize key presses and device state change events

class CEventHandler
{
public:
	virtual void OnKeyDown( int p_nKey ) = 0;				// Executed on each keydown event from device
	virtual void OnDeviceStateChanged( BOOL p_bState ) = 0; // True if device found, false if dissapeared
};

// **************** Support classes for command queuing and show picture requests buffering ****************

typedef struct _sSHOW_PICTURE_COMMAND {						// Images buffer structure
	BOOL m_bSent;
	int m_nFirstLine;
	int m_nLastLine;
	BYTE m_bBytes[ SCREEN_BUFFER_SIZE ];
	BYTE m_bDrawnBytes[ SCREEN_BUFFER_SIZE ];

#ifdef FEATURE_BLOCKING
	HANDLE m_hCompleteEvent;
	int *m_lpResult;
#endif

#ifdef FEATURE_LAYOUT
	int m_iLayoutState;
#endif 

	void Init()
	{
		m_bSent = TRUE;
#ifdef FEATURE_BLOCKING
		m_hCompleteEvent = NULL;
#endif

#ifdef FEATURE_LAYOUT
		m_iLayoutState = 0;
#endif
		memset( m_bDrawnBytes, 0, SCREEN_BUFFER_SIZE );
	}

	void Shown()
	{
		m_bSent = TRUE;
#ifdef FEATURE_BLOCKING
		m_hCompleteEvent = NULL;
#endif
		::CopyMemory( m_bDrawnBytes, m_bBytes, SCREEN_BUFFER_SIZE );
	}

} SHOW_PICTURE_COMMAND, *LPSHOW_PICTURE_COMMAND;

typedef struct _sOPTIMUS_COMMAND {							// Command queue structure
	BYTE m_c1, m_c2, m_c3, m_c4;

#ifdef FEATURE_BLOCKING
	HANDLE m_hCompleteEvent;
	int *m_lpResult;
#endif

} OPTIMUS_COMMAND, *LPOPTIMUS_COMMAND;

typedef std::queue<OPTIMUS_COMMAND> CCommandQueue;

#define MAX_EVENT 5       // init, close, command, overlap, close confirm
#define COMMAND_EVENT 2
#define OVERLAP_EVENT 3

// **************** OptimusMiniHelper class ****************

// Main class for Optimus mini usage. Thread safe.
// All device communication is in separate thread (as we should allways wait for keypress). 

class COptimusMiniHelper : public COptimusMini
{
protected:

	HANDLE m_hDeviceThread;				// thread for device communication
	DWORD m_dwDeviceThreadId;

#ifdef PORT_OVERLAPPED
	OVERLAPPED m_objOverlapped;			// for device thread overlapped operations
#endif
#ifdef FEATURE_LAYOUT
	BOOL m_bChangeLayout;
#endif

#ifdef FEATURE_REPAINT
	BOOL m_bRepaintPictures;
#endif 

	int m_nLastShownPicture;

	HANDLE m_hEvents[MAX_EVENT];

	CRITICAL_SECTION m_csPictureLocker;

	CCommandQueue m_qCommandQueue;

	CEventHandler* m_lpEventHandler;	// callbacks

	SHOW_PICTURE_COMMAND m_lpPictures[3]; // image buffer

public:

	COptimusMiniHelper();
	~COptimusMiniHelper();

	// Init device. Should be called initially and after WM_DEVICECHANGE
	// Calls OnDeviceStateChanged callback after successfull initialization
	BOOL Init();

	// Requests port close, blocking. Calls OnDeviceStateChanged callback
	void Close();

	void RegisterEventHandler( CEventHandler* p_lpEventHandler );

	// Adds command to device command queue
	int SendCommand( BYTE c1, BYTE c2, BOOL p_bBlock = FALSE );
	int SendCommand( BYTE c1, BYTE c2, BYTE c3, BYTE c4, BOOL p_bBlock = FALSE );
	
	// Adds picture to device picture bufffer
	int ShowPicture( int p_nButton, LPBYTE p_lpPictureBytes, BOOL p_bBlock = FALSE );

#ifdef FEATURE_LAYOUT
	// first downloading all new images, then showing them at once
	void ChangeLayout();
#endif

#ifdef FEATURE_REPAINT
	// requests all pictures repainting
	void RepaintPictures();
#endif

protected:
	// Executes commands from command queue (DEVICE THREAD)
	void SendCommandInternal();

	// Shos picture from device picture buffer (DEVICE THREAD)
	void ShowPictureInternal( int p_nButton );

	// Process keydown (DEVICE THREAD)
	void OnKeyDownInternal();

	// image buffer support function
	void FindPictureInternal();
	void ComparePictures( LPBYTE p_lpPictureA, LPBYTE p_lpPictureB, int* p_lpFirstLine, int* p_lpLastLine );

	// device communication thread main loop
	void DeviceThreadProcInternal();
	static DWORD WINAPI DeviceThreadProc( LPVOID lpParam );

	// calls OnKeyDown callback
	virtual void OnKeyDown( BYTE p_cValue );

	// calls OnDeviceStateChange callback
	virtual void OnDeviceStateChanged( BOOL p_bState );

};
